<* +m2extensions *>
IMPLEMENTATION MODULE Graphics;

IMPORT SYSTEM, Windows, Processes, Semaphores, Strings;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;

VAR
  createW, createH: INTEGER;
  W,H: INTEGER;
  scrollX,scrollY: INTEGER;
  bmap: Windows.HDC; (* device context to paint to *)
  hwnd: Windows.HWND; (* The window handle *)
  wprs: Processes.ProcessId;
  sem : Semaphores.SEMAPHORE;
  bmaplock: Semaphores.SEMAPHORE;

VAR
  brushtype: BrushType;
  brushcolor: COLOR;
  pentype: PenType;
  penwidth: CARDINAL;
  pencolor: COLOR;
  textcolor: COLOR;
  textfont: FONT;

CONST
  className = "XDSGraphicsWindowClss";
  windowStyle = Windows.WS_OVERLAPPEDWINDOW+Windows.WS_HSCROLL+Windows.WS_VSCROLL;
  defaultTitle = "Graphics";

CONST
  INPUTBUFSIZE = 32;

VAR
  inputCount, inputPut, inputGet: INTEGER;
  inputBuf: ARRAY [0..INPUTBUFSIZE] OF Input;
  inputWait: Semaphores.SEMAPHORE;

PROCEDURE initFonts; FORWARD;

(* ================= locks ================= *)

PROCEDURE lock;
BEGIN
  Semaphores.Claim(bmaplock);
END lock;

PROCEDURE unlock;
BEGIN
  Semaphores.Release(bmaplock);
END unlock;

(* ================= inputBufer ================= *)

PROCEDURE putInput(i: Input);
BEGIN
  lock;
    inputBuf[inputPut]:=i;
    inputPut:=(inputPut+1) MOD INPUTBUFSIZE;
    IF inputCount<INPUTBUFSIZE THEN
      INC(inputCount);
      Semaphores.Release(inputWait) (*   *)
    ELSE
      (*   ,     *)
      inputGet:=(inputGet+1) MOD INPUTBUFSIZE;
    END;
  unlock;
END putInput;

PROCEDURE getInput(VAR i: Input);
BEGIN
  Semaphores.Claim(inputWait); (*   *)
  lock;
    ASSERT(inputCount>0);
    i:=inputBuf[inputGet];
    inputGet:=(inputGet+1) MOD INPUTBUFSIZE;
    DEC(inputCount);
  unlock;
END getInput;

PROCEDURE inputs(): INTEGER;
BEGIN
  RETURN inputCount;
END inputs;

(* ================= pen and brush ================= *)

PROCEDURE createPen(type: PenType; width: CARDINAL; color: COLOR): Windows.HPEN;
  VAR desc: Windows.LOGPEN;
BEGIN
  desc.lopnWidth.x:=width;
  desc.lopnColor:=color;
  CASE type OF
    |pNull      : desc.lopnStyle:=Windows.PS_NULL;
    |pSolid     : desc.lopnStyle:=Windows.PS_SOLID;
    |pDot       : desc.lopnStyle:=Windows.PS_DOT;
    |pDash      : desc.lopnStyle:=Windows.PS_DASH;
    |pDashDot   : desc.lopnStyle:=Windows.PS_DASHDOT;
    |pDashDotDot: desc.lopnStyle:=Windows.PS_DASHDOTDOT;
  END;
  RETURN Windows.CreatePenIndirect(desc);
END createPen;

PROCEDURE createBrush(type: BrushType; color: COLOR): Windows.HBRUSH;
  VAR brush: Windows.HBRUSH;

  PROCEDURE makenone;
    VAR desc: Windows.LOGBRUSH;
  BEGIN
    desc.lbStyle:=Windows.BS_NULL;
    desc.lbColor:=0;
    desc.lbHatch:=0;
    brush:=Windows.CreateBrushIndirect(desc);
  END makenone;

  PROCEDURE makesolid(c: COLOR);
  BEGIN
    brush:=Windows.CreateSolidBrush(c);
  END makesolid;

  PROCEDURE makehatch(type: Windows.HS_ENUM; c: COLOR);
  BEGIN
    brush:=Windows.CreateHatchBrush(type,c);
  END makehatch;

BEGIN
  CASE type OF
    |bNull : makenone;
    |bSolid: makesolid(color);
    |bDiagb: makehatch(Windows.HS_BDIAGONAL,color);
    |bDiagf: makehatch(Windows.HS_FDIAGONAL,color);
    |bDiagcross: makehatch(Windows.HS_DIAGCROSS,color);
    |bVert : makehatch(Windows.HS_VERTICAL,color);
    |bHorz : makehatch(Windows.HS_HORIZONTAL,color);
    |bCross: makehatch(Windows.HS_CROSS,color);
  END;

  RETURN brush;
END createBrush;

(* ================= graphics window ================= *)

PROCEDURE update(x,y,w,h: INTEGER);
  VAR r: Windows.RECT;
BEGIN
  IF hwnd=NIL THEN HALT END;
  r.left:=x-1-scrollX; r.top:=y-1-scrollY; r.right:=x+w+1; r.bottom:=y+h+1;
  Windows.InvalidateRect(hwnd,r,TRUE);
END update;

PROCEDURE minmaxinfo(info: Windows.PMINMAXINFO);
  VAR r: Windows.RECT;
BEGIN
  r.left:=0; r.top:=0;
  r.right:=W; r.bottom:=H;
  Windows.AdjustWindowRect(r,windowStyle,FALSE);
  r.right:=r.right-r.left+Windows.GetSystemMetrics(Windows.SM_CXVSCROLL);
  r.bottom:=r.bottom-r.top+Windows.GetSystemMetrics(Windows.SM_CXHSCROLL);
  info^.ptMaxSize.x:=r.right;
  info^.ptMaxSize.y:=r.bottom;
  info^.ptMaxTrackSize.x:=r.right;
  info^.ptMaxTrackSize.y:=r.bottom;
END minmaxinfo;

PROCEDURE updateScrollBars(handle: Windows.HWND);
  VAR r: Windows.RECT;
      si: Windows.SCROLLINFO;
BEGIN
  Windows.GetClientRect(handle,r);
  si.cbSize:=SIZE(si);
  si.fMask:=Windows.SIF_RANGE+Windows.SIF_PAGE+Windows.SIF_POS;
  si.nMin:=0;
  si.nMax:=W;
  si.nPage:=r.right;
  si.nPos:=scrollX;
  Windows.SetScrollInfo(handle,Windows.SB_HORZ,si,TRUE);
  si.nMin:=0;
  si.nMax:=H;
  si.nPage:=r.bottom;
  si.nPos:=scrollY;
  Windows.SetScrollInfo(handle,Windows.SB_VERT,si,TRUE);
END updateScrollBars;

PROCEDURE handleHScroll(handle: Windows.HWND; code: Windows.SB_ENUM);
  VAR newpos: INTEGER;
      client: Windows.RECT;
      step: INTEGER;
      info: Windows.SCROLLINFO;
BEGIN
  Windows.GetClientRect(handle,client);
  step:=W DIV 100; IF step<=0 THEN step:=1 END;
  info.cbSize := SIZE(info);
  info.fMask  := Windows.SIF_POS+Windows.SIF_TRACKPOS;
  Windows.GetScrollInfo(handle,Windows.SB_HORZ,info);

  CASE code OF
    | Windows.SB_BOTTOM: newpos:=W-client.right;
    | Windows.SB_TOP   : newpos:=0;
    | Windows.SB_LINELEFT:
        newpos:=scrollX-step;
	IF newpos<0 THEN newpos:=0 END;
    | Windows.SB_LINERIGHT    (*, Windows.SB_LINEDOWN  *):
        newpos:=scrollX+step;
	IF newpos+client.right>W THEN newpos:=W-client.right END;
    | Windows.SB_PAGELEFT:
        newpos:=scrollX-client.right+1;
	IF newpos<0 THEN newpos:=0 END;
    | Windows.SB_PAGERIGHT:
        newpos:=scrollX+client.right-1;
	IF newpos+client.right>W THEN newpos:=W-client.right END;
    | Windows.SB_THUMBPOSITION:
        newpos:=info.nPos;
    | Windows.SB_THUMBTRACK:
        newpos:=info.nTrackPos;
  ELSE
    RETURN
  END;
  IF scrollX#newpos THEN
    scrollX:=newpos;
    Windows.InvalidateRect(handle,client,TRUE);
    updateScrollBars(handle);
  END;
END handleHScroll;

PROCEDURE handleVScroll(handle: Windows.HWND; code: Windows.SB_ENUM);
  VAR newpos: INTEGER;
      client: Windows.RECT;
      step: INTEGER;
      info: Windows.SCROLLINFO;
BEGIN
  Windows.GetClientRect(handle,client);
  step:=H DIV 100; IF step<=0 THEN step:=1 END;
  info.cbSize := SIZE(info);
  info.fMask  := Windows.SIF_POS+Windows.SIF_TRACKPOS;
  Windows.GetScrollInfo(handle,Windows.SB_VERT,info);

  CASE code OF
    | Windows.SB_BOTTOM: newpos:=H-client.bottom;
    | Windows.SB_TOP   : newpos:=0;
    | Windows.SB_LINEUP:
        newpos:=scrollY-step;
	IF newpos<0 THEN newpos:=0 END;
    | Windows.SB_LINEDOWN:
        newpos:=scrollY+step;
	IF newpos+client.bottom>H THEN newpos:=H-client.bottom END;
    | Windows.SB_PAGEUP:
        newpos:=scrollY-client.bottom+1;
	IF newpos<0 THEN newpos:=0 END;
    | Windows.SB_PAGEDOWN:
        newpos:=scrollY+client.bottom-1;
	IF newpos+client.bottom>H THEN newpos:=H-client.bottom END;
    | Windows.SB_THUMBPOSITION:
        newpos:=info.nPos;
    | Windows.SB_THUMBTRACK:
        newpos:=info.nTrackPos;
  ELSE
    RETURN
  END;
  IF scrollY#newpos THEN
    scrollY:=newpos;
    Windows.InvalidateRect(handle,client,TRUE);
    updateScrollBars(handle);
  END;
END handleVScroll;

PROCEDURE paint(hwnd: Windows.HWND);
  VAR
    hdc: Windows.HDC;
    x,y,w,h: INTEGER;
    ps: Windows.PAINTSTRUCT;
BEGIN
  hdc := Windows.BeginPaint (hwnd, ps);
    x:=ps.rcPaint.left;
    y:=ps.rcPaint.top;
    w:=ps.rcPaint.right-ps.rcPaint.left;
    h:=ps.rcPaint.bottom-ps.rcPaint.top;
--    lock;
      Windows.BitBlt(hdc,x,y,w,h,
                     bmap,x+scrollX, y+scrollY,
                     Windows.SRCCOPY);
--    unlock;	
  Windows.EndPaint (hwnd, ps);
END paint;

CONST
  INIT = Input {keyUp, 0,0, Buttons{},mouseLeft,0 };
(*
    event: Event;
    x,y: INTEGER;   (* mouse coordinates in "picture", not in window *)
    state: Buttons;
    button: Button; (* for mouse event - mouseLeft or mouseRigt *)
    key: INTEGER;   (* for keyboard event - the pressed/releasded key code *)
*)

PROCEDURE mouseEvent(e: Event; b: Button; wParam: Windows.WPARAM; lParam: Windows.LPARAM);
  VAR i: Input;
      in: Windows.MK_SET;
BEGIN
  i:=INIT;
  i.event:=e; i.button:=b;
  i.x:=Windows.LOWORD(lParam)+scrollX;
  i.y:=Windows.HIWORD(lParam)+scrollY;

  in:=SYSTEM.CAST(Windows.MK_SET,wParam);
  IF Windows.MK_SHIFT*in#Windows.MK_SET{} THEN
    INCL(i.state,shift);
  END;
  IF Windows.MK_CONTROL*in#Windows.MK_SET{} THEN
    INCL(i.state,control);
  END;
  IF Windows.MK_LBUTTON*in#Windows.MK_SET{} THEN
    INCL(i.state,mouseLeft);
  END;
  IF Windows.MK_MBUTTON*in#Windows.MK_SET{} THEN
    INCL(i.state,mouseMiddle);
  END;
  IF Windows.MK_RBUTTON*in#Windows.MK_SET{} THEN
    INCL(i.state,mouseRight);
  END;
  putInput(i);
END mouseEvent;

PROCEDURE keyboardEvent(e: Event; wParam: Windows.WPARAM; lParam: Windows.LPARAM);
  VAR i: Input;
   kstate: SYSTEM.INT16;
BEGIN
  i:=INIT;
  i.event:=e;
  i.key:=wParam;

  i.state:=Buttons{};
  kstate:=Windows.GetKeyState(Windows.VK_SHIFT);
  IF kstate<0 THEN INCL(i.state,shift)   END;
  kstate:=Windows.GetKeyState(Windows.VK_CONTROL);
  IF kstate<0 THEN INCL(i.state,control) END;
  kstate:=Windows.GetKeyState(Windows.VK_MENU);
  IF kstate<0 THEN INCL(i.state,alt) END;
  kstate:=Windows.GetKeyState(Windows.VK_LBUTTON);
  IF kstate<0 THEN INCL(i.state,mouseLeft) END;
  kstate:=Windows.GetKeyState(Windows.VK_RBUTTON);
  IF kstate<0 THEN INCL(i.state,mouseRight) END;
  kstate:=Windows.GetKeyState(Windows.VK_MBUTTON);
  IF kstate<0 THEN INCL(i.state,mouseMiddle) END;
  putInput(i);
END keyboardEvent;

PROCEDURE [Windows.CALLBACK] wndProc (
  handle : Windows.HWND;
  message: Windows.UINT;
  wParam : Windows.WPARAM;
  lParam : Windows.LPARAM) : Windows.LRESULT;
BEGIN
  CASE message OF
   |Windows.WM_CREATE:
      updateScrollBars(handle);
   |Windows.WM_SIZE:
      updateScrollBars(handle);
   |Windows.WM_HSCROLL:
      handleHScroll(handle,SYSTEM.CAST(Windows.SB_ENUM,Windows.LOWORD(wParam)));
   |Windows.WM_VSCROLL:
      handleVScroll(handle,SYSTEM.CAST(Windows.SB_ENUM,Windows.LOWORD(wParam)));
   |Windows.WM_PAINT:
      lock;
        paint(handle);
      unlock;
   |Windows.WM_CLOSE:
      Windows.PostQuitMessage(0);
   |Windows.WM_GETMINMAXINFO:
      minmaxinfo(SYSTEM.CAST(Windows.PMINMAXINFO,lParam));
   |Windows.WM_LBUTTONDOWN:
      mouseEvent(mouseDown,mouseLeft,wParam,lParam);
   |Windows.WM_LBUTTONUP:
      mouseEvent(mouseUp,mouseLeft,wParam,lParam);
   |Windows.WM_LBUTTONDBLCLK:
      mouseEvent(mouseDblClick,mouseLeft,wParam,lParam);
   |Windows.WM_RBUTTONDOWN:
      mouseEvent(mouseDown,mouseRight,wParam,lParam);
   |Windows.WM_RBUTTONUP:
      mouseEvent(mouseUp,mouseRight,wParam,lParam);
   |Windows.WM_RBUTTONDBLCLK:
      mouseEvent(mouseDblClick,mouseRight,wParam,lParam);
   |Windows.WM_MBUTTONDOWN:
      mouseEvent(mouseDown,mouseMiddle,wParam,lParam);
   |Windows.WM_MBUTTONUP:
      mouseEvent(mouseUp,mouseMiddle,wParam,lParam);
   |Windows.WM_MBUTTONDBLCLK:
      mouseEvent(mouseDblClick,mouseMiddle,wParam,lParam);
   |Windows.WM_KEYDOWN:
      keyboardEvent(keyDown,wParam,lParam);
   |Windows.WM_KEYUP:
      keyboardEvent(keyUp,wParam,lParam);
  ELSE
    RETURN Windows.DefWindowProc(handle,message,wParam,lParam);
  END;
  RETURN 0;
END wndProc;

PROCEDURE registerClass(name: ARRAY OF CHAR);
  VAR wc : Windows.WNDCLASSEX;
BEGIN
    wc.cbSize  := SIZE (Windows.WNDCLASSEX);

    wc.style         := Windows.CS_DBLCLKS;
    wc.lpfnWndProc   := wndProc;
    wc.cbClsExtra    := 0;
    wc.cbWndExtra    := 0;
    wc.hInstance     := Windows.GetModuleHandle(NIL);
    wc.hCursor       := NIL;
    wc.hbrBackground := NIL;
    wc.lpszMenuName  := NIL;
    wc.lpszClassName := SYSTEM.CAST(Windows.PSTR,SYSTEM.ADR (name));

    wc.hIcon         := NIL;
    wc.hIconSm       := NIL;

    ASSERT(Windows.RegisterClassEx (wc) # 0);
END registerClass;

PROCEDURE createBimap;
  VAR
    desktop: Windows.HDC;
    bmp: Windows.HBITMAP;
    brush: Windows.HBRUSH;
    rect: Windows.RECT;

BEGIN
  desktop:=Windows.GetDC(Windows.GetDesktopWindow());
  bmp    := Windows.CreateCompatibleBitmap (desktop,createW, createH);
  bmap   := Windows.CreateCompatibleDC (desktop);
  Windows.SelectObject (bmap, bmp);
  W:=createW; H:=createH;

  brush:=createBrush(bSolid,0FFFFFFH);
  Windows.SelectObject(bmap,brush);
  brushtype:=bSolid; brushcolor:=0FFFFFFH;
  Windows.SelectObject(bmap,createPen(pSolid,1,0));
  pentype:=pSolid; penwidth:=1; pencolor:=0;

  rect.left:=0;    rect.top:=0;
  rect.right:=W; rect.bottom:=H;
  Windows.FillRect(bmap,rect,brush);
  Windows.SetBkMode   (bmap, Windows.TRANSPARENT);

  initFonts;
END createBimap;

PROCEDURE createWindow;
BEGIN
  createBimap;
  hwnd:=Windows.CreateWindow(
    className,defaultTitle,windowStyle,
    Windows.CW_USEDEFAULT,0,Windows.CW_USEDEFAULT,0,
    NIL,NIL,NIL,NIL);
  Windows.ShowWindow   (hwnd, Windows.SW_SHOWDEFAULT);
  Windows.UpdateWindow (hwnd);
END createWindow;

PROCEDURE windowprocess;
  VAR msg: Windows.MSG;
BEGIN
  createWindow;
  Semaphores.Release(sem);
  WHILE Windows.GetMessage (msg, NIL, 0, 0) DO
    Windows.TranslateMessage (msg);
    Windows.DispatchMessage  (msg);
  END;
  hwnd:=NIL;
  Semaphores.Release(sem);
END windowprocess;

(* ================= painting ================= *)

PROCEDURE RGB(r,g,b: CARD8): COLOR;
BEGIN RETURN b*256*256+g*256+r END RGB;

PROCEDURE SetBrush(type: BrushType; color: COLOR);
  VAR old: Windows.HBRUSH;
BEGIN
  lock;
    old:=Windows.SelectObject(bmap,createBrush(type,color));
  unlock;
  Windows.DeleteObject(old);
  brushtype:=type; brushcolor:=color;
END SetBrush;

PROCEDURE GetBrush(VAR type: BrushType; VAR color: COLOR);
BEGIN
  type:=brushtype;
  color:=brushcolor;
END GetBrush;

PROCEDURE SetSolidPen(width: CARDINAL; color: COLOR);
  VAR old: Windows.HPEN;
BEGIN
  lock;
    old:=Windows.SelectObject(bmap,createPen(pSolid,width,color));
  unlock;
  Windows.DeleteObject(old);
  pentype:=pSolid; penwidth:=width; pencolor:=color;
END SetSolidPen;

PROCEDURE SetNullPen();
  VAR old: Windows.HPEN;
BEGIN
  lock;
    old:=Windows.SelectObject(bmap,createPen(pNull,1,0));
  unlock;
  Windows.DeleteObject(old);
  pentype:=pNull; penwidth:=1; pencolor:=0;
END SetNullPen;

PROCEDURE SetStylePen(type: PenType; color: COLOR);
  VAR old: Windows.HPEN;
BEGIN
  lock;
    old:=Windows.SelectObject(bmap,createPen(type,1,color));
  unlock;
  Windows.DeleteObject(old);
  pentype:=type; penwidth:=1; pencolor:=color;
END SetStylePen;

PROCEDURE GetPen(VAR type: PenType; VAR width: CARDINAL; VAR color: COLOR);
BEGIN type:=pentype; width:=penwidth; color:=pencolor END GetPen;

PROCEDURE Line(x0,y0,x1,y1: INTEGER);
  VAR w,h: INTEGER;
BEGIN
  lock;
    Windows.MoveToEx(bmap,x0,y0,NIL);
    Windows.LineTo(bmap,x1,y1);
  unlock;
  w:=x1-x0; IF w<0 THEN w:=-w; x0:=x1 END;
  h:=y1-y0; IF h<0 THEN h:=-h; y0:=y1 END;
  update(x0,y0,w,h);
END Line;

PROCEDURE SolidLine(x0,y0,x1,y1: INTEGER; width: INTEGER; c: COLOR);
  VAR old,new: Windows.HPEN;
      w,h: INTEGER;
BEGIN
  new:=Windows.CreatePen(Windows.PS_SOLID,width,c);
  lock;
    old:=Windows.SelectObject(bmap,new);
    Windows.MoveToEx(bmap,x0,y0,NIL);
    Windows.LineTo(bmap,x1,y1);
    old:=Windows.SelectObject(bmap,old);
  unlock;
  Windows.DeleteObject(new);
  w:=x1-x0; IF w<0 THEN w:=-w; x0:=x1 END;
  h:=y1-y0; IF h<0 THEN h:=-h; y0:=y1 END;
  update(x0,y0,w,h);
END SolidLine;

PROCEDURE Rectangle(x,y,w,h: INTEGER);
BEGIN
  lock;
    Windows.Rectangle(bmap,x,y,x+w,y+h);
  unlock;
  update(x,y,w,h);
END Rectangle;

PROCEDURE FilledRect(x,y,w,h: INTEGER; c: COLOR);
  VAR rect: Windows.RECT;
      brush: Windows.HBRUSH;
BEGIN
  rect.left:=x;    rect.top:=y;
  rect.right:=x+w; rect.bottom:=y+h;
  brush:=Windows.CreateSolidBrush(c);

  lock;
    Windows.FillRect(bmap,rect,brush);
  unlock;

  Windows.DeleteObject(brush);
  update(x,y,w,h);
END FilledRect;

PROCEDURE OutlineRect(x,y,w,h: INTEGER; linewidth: INTEGER; c: COLOR);
  VAR rect: Windows.RECT;
      brush,oldbrush: Windows.HBRUSH;
      pen,oldpen: Windows.HPEN;
BEGIN
  rect.left:=x;    rect.top:=y;
  rect.right:=x+w; rect.bottom:=y+h;
  brush:=createBrush(bNull,0);
  pen:=createPen(pSolid,linewidth,c);

  lock;
    oldbrush:=Windows.SelectObject(bmap,brush);
    oldpen:=Windows.SelectObject(bmap,pen);
    Windows.Rectangle(bmap,x,y,x+w,y+h);
    oldbrush:=Windows.SelectObject(bmap,oldbrush);
    oldpen:=Windows.SelectObject(bmap,oldpen);
  unlock;

  Windows.DeleteObject(brush);
  Windows.DeleteObject(pen);
  update(x,y,w,h);
END OutlineRect;

PROCEDURE RoundRect(x,y,w,h,r: INTEGER);
BEGIN
  lock;
    Windows.RoundRect(bmap,x,y,x+w,y+h,r,r);
  unlock;
  update(x,y,w,h);
END RoundRect;

PROCEDURE Ellipse(x,y,w,h: INTEGER);
BEGIN
  lock;
    Windows.Ellipse(bmap,x,y,x+w,y+h);
  unlock;
  update(x,y,w,h);
END Ellipse;

(* ================= mouse ================= *)
TYPE FONT = POINTER TO Windows.LOGFONTA;

PROCEDURE NewFont(name: ARRAY OF CHAR; height: CARDINAL; bold,italic: BOOLEAN): FONT;
(* returns NIL if failure *)
  VAR new: FONT;
BEGIN
  NEW(new);
  SYSTEM.FILL(new,0,SIZE(new^));
  new^.lfHeight := -VAL(INTEGER,height);
  IF bold THEN new^.lfWeight:=Windows.FW_BOLD ELSE new^.lfWeight:=Windows.FW_NORMAL END;
  new^.lfItalic:=italic;
  new^.lfCharSet:=Windows.DEFAULT_CHARSET;
  Strings.Assign(name,new^.lfFaceName);
  RETURN new;
END NewFont;

PROCEDURE DelFont(VAR f: FONT);
(* delenes font and sets NIL to f *)
BEGIN
  DISPOSE(f); f:=NIL;
END DelFont;

PROCEDURE SetFont(f: FONT);
(* sets the current font *)
  VAR font: Windows.HFONT;
BEGIN
  font:=Windows.CreateFontIndirectA(f^);
  lock;
    font:=Windows.SelectObject(bmap,font);
  unlock;
  Windows.DeleteObject(font);
  textfont:=f;
END SetFont;

PROCEDURE GetFont(): FONT;
BEGIN
  RETURN textfont;
END GetFont;

PROCEDURE SetTextColor(c: COLOR);
BEGIN
  lock;
    Windows.SetTextColor(bmap, c);
  unlock;
  textcolor:=c;
END SetTextColor;

PROCEDURE GetTextColor(): COLOR;
BEGIN
  RETURN textcolor;
END GetTextColor;

PROCEDURE WriteString(x,y: INTEGER; s: ARRAY OF CHAR);
  VAR rect: Windows.RECT; size: Windows.SIZEL;
      slen: INTEGER;
BEGIN
  rect.left:=x; rect.top:=y;
  rect.bottom:=GetH(); rect.right:=GetW();
  slen:=Strings.Length(s);

  lock;
    Windows.DrawText    (bmap, s, slen, rect,
                         Windows.DT_LEFT + Windows.DT_TOP + Windows.DT_SINGLELINE);
    Windows.GetTextExtentPoint32(bmap,s,slen,size);
  unlock;
  update(x,y,size.cx,size.cy);
END WriteString;

PROCEDURE StringSize(s: ARRAY OF CHAR; VAR w,h: INTEGER);
  VAR size: Windows.SIZEL;
BEGIN
  lock;
    Windows.GetTextExtentPoint32(bmap,s,Strings.Length(s),size);
  unlock;
  w:=size.cx; h:=size.cy;
END StringSize;

PROCEDURE initFonts;
BEGIN
  arial  :=NewFont("Arial"      ,0,FALSE,FALSE);
  courier:=NewFont("Courier New",0,FALSE,FALSE);
  times  :=NewFont("Times"      ,0,FALSE,FALSE);
  SetFont(arial);
END initFonts;

(* ================= input ================= *)

PROCEDURE IsInputReady(): BOOLEAN;
BEGIN
  RETURN (inputs()>0);
END IsInputReady;

PROCEDURE ReadInput(VAR i: Input);
BEGIN
  getInput(i);
END ReadInput;

PROCEDURE PeekInput(VAR inp: Input): BOOLEAN;
BEGIN
  IF IsInputReady() THEN
    ReadInput(inp); RETURN TRUE
  END;
  RETURN FALSE
END PeekInput;


PROCEDURE WaitForClick();
  VAR i: Input;
BEGIN
  ASSERT(hwnd#NIL);
  REPEAT
    ReadInput(i);
  UNTIL (hwnd=NIL) OR ((i.event=mouseDown) & (i.button=mouseLeft));
END WaitForClick;

(* ================= general control procedures ================= *)

PROCEDURE SetWindowTitle(title: ARRAY OF CHAR);
(* ENG: set grahics window title a given text. Use emty string to set the default title.
   RUS:      .   
     ( ) *)
BEGIN
  IF title[0]=0C THEN
    Windows.SetWindowText(hwnd,defaultTitle);
  ELSE
    Windows.SetWindowText(hwnd,title);
  END;
END SetWindowTitle;

PROCEDURE SetSize(width,height: INTEGER);
BEGIN
  createW:=width;
  createH:=height;
END SetSize;

PROCEDURE GetW(): INTEGER; BEGIN RETURN W END GetW;

PROCEDURE GetH(): INTEGER; BEGIN RETURN H END GetH;

PROCEDURE OpenGraphics;
BEGIN
  Semaphores.Create(sem,0);
  Semaphores.Create(bmaplock,1);
  Semaphores.Create(inputWait,0);
  inputCount:=0; inputPut:=0; inputGet:=0;
  Processes.Start(windowprocess,0,0,NIL,wprs);
  Semaphores.Claim(sem);
END OpenGraphics;

BEGIN
  W:=-1; H:=-1;
  scrollX:=0; scrollY:=0;
  createW:=800; createH:=600;
  registerClass(className);
FINALLY
  IF Semaphores.CondClaim(sem) THEN
    Windows.SetWindowText(hwnd,"     ");
    Semaphores.Claim(sem);
  END;
END Graphics.